from .helpers import get_last_index_of_substring
from os import path, remove, rename
from shutil import rmtree, move
from pathlib import Path


def perform_adds(paths):
    """Creates all files from given paths

    Args:
        paths (array): Array of string paths

    """
    for new_path in paths:
        last_slash_index = get_last_index_of_substring(new_path, "/")
        directory = new_path[: last_slash_index + 1]
        filename = new_path[last_slash_index + 1 :]

        p = Path(directory)
        p.mkdir(parents=True, exist_ok=True)

        if filename:
            open(p / filename, 'a').close()


def perform_deletes(paths):
    """Deletes all files from given paths
    If the path doesn't exist, bmv will silently continue
    with next operations

    Args:
        paths (array): Array of string paths

    """
    for path_to_delete in paths:
        if path.exists(path_to_delete):
            if path.isdir(path_to_delete):
                rmtree(path_to_delete)
            elif path.isfile(path_to_delete):
                remove(path_to_delete)


def perform_renames(renamings):
    """Renames old file/dir names to new ones as
    specified by renamings

    Args:
        renamings (dict): dict containing old_path and new_path as keys

    """
    for renaming in renamings:
        old_name = renaming["old_path"]
        new_name = renaming["new_path"]
        print(old_name, new_name)
        rename(old_name, new_name)


def perform_moves(movements):
    """Moves old file/dir names to new paths as
    specified by movements

    Args:
        movements (dict): dict containing current_path and new_path as keys

    """
    for movement in movements:
        current_path = movement["current_path"]
        new_path = movement["new_path"]
        move(current_path, new_path)



import unittest
from bulk_mv import build_from_directory, parse_bmv, BmvGenerator
from pathlib import Path


class TestBmvGenerator(unittest.TestCase):
    def test_generator(self):
        self.maxDiff = 10000
        file_tree = build_from_directory(f"{Path(__file__).parent}/dummy_directories/sample1/")
        bmv_content = BmvGenerator().generate(file_tree)

        expected_output = (
            "[" + str(Path(__file__).parent) + "/dummy_directories/sample1] {\n"
            "\tindex.html\n"
            "\t[admin] {\n"
            "\t\tindex.html\n"
            "\t\tnichoal.html\n"
            "\t}\n"
            "\t[users] {\n"
            "\t\tangel.html\n"
            "\t\tindex.html\n"
            "\t\tjosh.html\n"
            "\t\t[blocked] {\n"
            "\t\t\tblocked_user_1.html\n"
            "\t\t}\n"
            "\t}\n"
            "}\n"
        )

        self.assertEqual(expected_output, bmv_content)



def get_last_index_of_substring(string, substring):
    """Returns index of the final substring occurring in a given string
    For example, the index of the last '/' character in
    path/to/some/file is 12.

    Args:
        string (str): The full string
        substring (str): The substring your searching for in string

    Returns:
        int: the last index, or -1 if the substring is not found
    """

    try:
        last_index = len(string) - 1 - string[::-1].index(substring)
        return last_index
    except ValueError:
        return -1



import unittest
import random
from bulk_mv import build_from_directory, FileTree
from pathlib import Path
import os


class TestFileTree(unittest.TestCase):
    def build_sample_1(self):
        sample1 = build_from_directory(f"./bulk_mv/tests/dummy_directories/sample1/")
        return sample1

    def build_sample_2(self):
        sample2 = build_from_directory(f"./bulk_mv/tests/dummy_directories/sample2/")
        return sample2

    def build_sample_3(self):
        sample3 = build_from_directory(f"./bulk_mv/tests/dummy_directories/sample3/")
        return sample3

    def test_all_paths(self):
        sample3 = self.build_sample_3()
        expected_output = [
            './bulk_mv/tests/dummy_directories/sample3/',
            './bulk_mv/tests/dummy_directories/sample3/photos/',
            './bulk_mv/tests/dummy_directories/sample3/photos/test.jpg',
            './bulk_mv/tests/dummy_directories/sample3/markdown/',
            './bulk_mv/tests/dummy_directories/sample3/markdown/1.md',
            './bulk_mv/tests/dummy_directories/sample3/markdown/2.md',
            './bulk_mv/tests/dummy_directories/sample3/delete/',
            './bulk_mv/tests/dummy_directories/sample3/delete/no_more.html',
            './bulk_mv/tests/dummy_directories/sample3/text/',
            './bulk_mv/tests/dummy_directories/sample3/text/1.txt',
            './bulk_mv/tests/dummy_directories/sample3/text/2.txt',
        ]

        output = sample3.all_paths()

        self.assertSetEqual(set(output), set(expected_output))

    def test_name(self):
        dir = FileTree("dir")
        dir.add_child(FileTree("file.txt"))

        sub_dir = FileTree("sub_dir")
        sub_dir.add_child(FileTree("test.txt"))
        dir.add_child(sub_dir)

        self.assertEqual(dir.name(), "dir")
        self.assertEqual(dir.children()[0].name(), "file.txt")
        self.assertEqual(dir.children()[1].name(), "sub_dir")
        self.assertEqual(dir.children()[1].children()[0].name(), "test.txt")

    def test_add_child(self):
        dir = FileTree("dir")
        sub_dir = FileTree("sub_dir")

        file_txt = FileTree("file.txt")
        test_txt = FileTree("test.txt")

        dir.add_child(file_txt)
        sub_dir.add_child(test_txt)
        dir.add_child(sub_dir)

        self.assertEqual(dir._path(), "dir")
        self.assertEqual(dir.children()[0]._path(), "dir/file.txt")
        self.assertEqual(dir.children()[1]._path(), "dir/sub_dir")
        self.assertEqual(dir.children()[1].children()[0]._path(), "dir/sub_dir/test.txt")
        self.assertEqual(sub_dir._parent(), dir)
        self.assertEqual(file_txt._parent(), dir)
        self.assertEqual(test_txt._parent(), sub_dir)

    def test_inequality(self):
        sample1 = self.build_sample_1()
        sample2 = self.build_sample_2()

        self.assertNotEqual(sample1, sample2)

    def test_exact_equality(self):
        sample1 = self.build_sample_1()
        sample1_dup = self.build_sample_1()

        self.assertEqual(sample1, sample1_dup)

    def test_effective_equality(self):
        # What this means is suppose you have two directories A and B
        # (imagine A and B have the same name). If A has files: 1.txt, 2.txt and 3.txt
        # and B has files: 3.txt, 2.txt and 1.txt, A and B are equivalent. The order
        # in which the files are presented should not matter.

        sample1 = self.build_sample_1()
        sample1_dup = self.build_sample_1()

        stack = [sample1_dup]
        while stack:
            tree = stack.pop()
            random.shuffle(tree.children())

            for child in tree.children():
                if child._is_dir():
                    stack.append(child)

        self.assertEqual(sample1, sample1_dup)

    def test_extra_directory(self):
        sample1 = self.build_sample_1()
        ft = self.build_sample_1()

        dir = FileTree("some_dir")
        dir.add_child(FileTree("a.txt"))
        dir.add_child(FileTree("b.txt"))
        ft.add_child(dir)

        self.assertNotEqual(sample1, ft)

    def test_extra_file(self):
        sample1 = self.build_sample_1()
        ft = self.build_sample_1()
        ft.add_child(FileTree("some_file"))

        self.assertNotEqual(sample1, ft)



import unittest
from bulk_mv import (
    build_from_directory,
    parse_bmv,
    perform_adds,
    perform_deletes,
    perform_renames,
    perform_moves,
)
from os import path, remove, rename
from shutil import rmtree, move
import pathlib


class TestOperations(unittest.TestCase):
    def setUp(self):
        with open("bulk_mv/tests/dummy_bmv_files/sample3.bmv", "r") as file:
            self.output = parse_bmv(file.read().strip())

    def test_perform_adds(self):
        perform_adds(self.output["add"])

        self.assertTrue(path.exists("./bulk_mv/tests/dummy_directories/sample3/new_file.txt"))

        remove("./bulk_mv/tests/dummy_directories/sample3/new_file.txt")

    def test_perform_deletes(self):
        perform_deletes(self.output["delete"])

        self.assertTrue(not path.exists("./bulk_mv/tests/dummy_directories/sample3/markdown/2.md"))
        self.assertTrue(not path.exists("./bulk_mv/tests/dummy_directories/sample3/delete/"))

        open("./bulk_mv/tests/dummy_directories/sample3/markdown/2.md", 'a').close()
        pathlib.Path("./bulk_mv/tests/dummy_directories/sample3/delete/").mkdir(parents=True, exist_ok=True)
        open("./bulk_mv/tests/dummy_directories/sample3/delete/no_more.html", 'a').close()

    def test_perform_renames(self):
        perform_renames(self.output["rename_files"])
        perform_renames(self.output["rename_dirs"])

        self.assertTrue(path.exists("./bulk_mv/tests/dummy_directories/sample3/txt/"))

        rename("./bulk_mv/tests/dummy_directories/sample3/txt/", "./bulk_mv/tests/dummy_directories/sample3/text/")

    def test_perform_moves(self):
        perform_moves(self.output["move_files"])
        perform_moves(self.output["move_dirs"])

        self.assertTrue(path.exists("./bulk_mv/tests/dummy_directories/sample3/photos/markdown/"))

        move(
            "./bulk_mv/tests/dummy_directories/sample3/photos/markdown/",
            "./bulk_mv/tests/dummy_directories/sample3//markdown",
        )



from lark import Lark, Transformer
from .helpers import get_last_index_of_substring


class TreeToOperations(Transformer):
    def start(self, items):
        return items[0]

    def dir(self, items):
        """Creates a dict representation of a directory.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a directory
        """
        dirname = items[1]['dir_op']['dirname']
        dirs = []
        files = []
        unary_ops = []
        file_op = []
        dir_op = []

        info_offset = 6
        if 'files' in items[info_offset]:
            files = items[info_offset]['files']
        if 'dirs' in items[info_offset]:
            dirs = items[info_offset]['dirs']
        if 'unary' in items[info_offset]:
            unary_ops = items[info_offset]['unary']
        if 'file_op' in items[info_offset]:
            for op in items[info_offset]['file_op']:
                if 'file' in op:
                    files.append(op['file'].value)
                elif 'rename_file_op' in op:
                    file_op.append(op)
                elif 'move_file_op' in op:
                    file_op.append(op)

        if 'move' in items[1]['dir_op']:
            dir_op.append({'move': items[1]['dir_op']['move']})
        if 'rename' in items[1]['dir_op']:
            dir_op.append({'rename': items[1]['dir_op']['rename']})

        return {
            'dir': {
                'dirname': dirname,
                'files': files,
                'unary': unary_ops,
                'dir_op': dir_op,
                'file_op': file_op,
                'dirs': dirs,
            }
        }

    def dir_op(self, items):
        """Creates a dict representation of a directory operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a directory operation
        """
        return {'dir_op': items[0]}

    def rename_dir_op(self, items):
        """Creates a dict representation of a directory rename operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a directory rename operation
        """
        old_name = items[0]['dirname']
        new_name = items[4]['dirname']
        return {'dirname': old_name, 'rename': {'old_name': old_name, 'new_name': new_name}}

    def move_dir_op(self, items):
        """Creates a dict representation of a directory move operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a directory move operation
        """
        current_name = items[0]['dirname']
        new_path = items[4]['dirname']
        return {'dirname': current_name, 'move': {'current_name': current_name, 'new_path': new_path}}

    def block(self, items):
        """Creates a dict representation of a block.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a block
        """
        files = []
        dirs = []
        unary_ops = []
        file_op = []
        dir_op = []

        for item in items:
            if 'file' in item:
                files.append(item['file'])
            elif 'dir' in item:
                dirs.append(item['dir'])
                for op in item['dir']['dir_op']:
                    if 'move' in op:
                        dir_op.append(op)
            elif 'unary' in item:
                unary_ops.append(item['unary'])
            elif 'file_op' in item:
                if 'file' in item['file_op']:
                    files.append(item['file_op']['file'])
                elif 'rename_file_op' in item['file_op']:
                    file_op.append(item['file_op'])
                elif 'move_file_op' in item['file_op']:
                    file_op.append(item['file_op'])

        return {'dirs': dirs, 'files': files, 'unary': unary_ops, 'file_op': file_op}

    def block_item(self, items):
        """Creates a dict representation of a block item.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a block item
        """
        return items[0]

    def entity(self, items):
        """Creates a dict representation of an entity.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of an entity
        """
        if 'file' in items[0]:
            return {'file': items[0]['file']}

        return items[0]

    def file_op(self, items):
        """Creates a dict representation of a file operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a file operation
        """
        return {'file_op': items[0]}

    def rename_file_op(self, items):
        """Creates a dict representation of a file rename operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a file rename operation
        """
        old_file_name = items[0]['file']
        new_file_name = items[4]['file']
        return {'rename_file_op': {'old_name': old_file_name, 'new_name': new_file_name}}

    def move_file_op(self, items):
        """Creates a dict representation of a file move operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a file move operation
        """
        current_name = items[0]['file']
        new_path = items[4]['dirname']
        return {'move_file_op': {'current_name': current_name, 'new_path': new_path}}

    def unary_op(self, items):
        """Creates a dict representation of a unary operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a unary operation
        """
        return {'unary': items[0]}

    def add_op(self, items):
        """Creates a dict representation of an add operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of an add operation
        """
        return {'add': items[1]}

    def delete_op(self, items):
        """Creates a dict representation of a delete operation.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a delete operation
        """
        return {'delete': items[1]}

    def dirname(self, items):
        """Creates a dict representation of a directory name.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a directory name
        """
        return {'dirname': items[0].value}

    def filename(self, items):
        """Creates a dict representation of a file name.

        Args:
            items (array): Array of node items from the Lark parser

        Returns:
            dict: dict representation of a file name
        """
        return {'file': items[0].value}

    def rename_op(self, _):
        return ">"

    def lbrace(self, _):
        return "{"

    def rbrace(self, _):
        return "}"

    def lbrack(self, _):
        return "["

    def rbrack(self, _):
        return "]"


def get_operations(tree, ops, current_path):
    """Extracts operation instructions from BMV Parse tree

    Args:
        tree (dict): BMV Parse tree
        ops (dict): Dictionary to store operations
        current_path (str): Current path of the directory

    Returns:
        dict: Dictionary of operations
    """
    for op in tree['unary']:
        if 'add' in op:
            if 'file' in op['add']:
                final_path = current_path + '/' + op['add']['file']
                ops['add'].append(final_path.replace("//", "/"))
            elif 'dir' in op['add']:
                final_path = current_path + '/' + op['add']['dir']['dirname']
                ops['add'].append((final_path + '/').replace("//", "/"))

                for file in op['add']['dir']['files']:
                    ops['add'].append((final_path + '/' + file).replace("//", "/"))

                get_operations(op['add']['dir'], ops, final_path.replace("//", "/"))

        elif 'delete' in op:
            if 'file' in op['delete']:
                final_path = current_path + '/' + op['delete']['file']
                ops['delete'].append(final_path.replace("//", "/"))
            elif 'dir' in op['delete']:
                final_path = current_path + '/' + op['delete']['dir']['dirname']
                ops['delete'].append((final_path + '/').replace("//", "/"))

    for op in tree['file_op']:
        if "rename_file_op" in op:
            old_path = current_path + '/' + op['rename_file_op']['old_name']
            new_path = current_path + '/' + op['rename_file_op']['new_name']
            ops['rename_files'].append({'old_path': old_path, 'new_path': new_path})
        if "move_file_op" in op:
            old_path = current_path + '/' + op['move_file_op']['current_name']
            new_path = op['move_file_op']['new_path']
            ops['move_files'].append({'current_path': old_path, 'new_path': new_path})

    for op in tree['dir_op']:
        if 'rename' in op:
            if current_path != op['rename']['old_name']:
                old_path = current_path + '/'

                last_slash_index = get_last_index_of_substring(current_path, "/")
                new_path = current_path[:last_slash_index] + '/' + op['rename']['new_name']
                ops['rename_dirs'].append({'old_path': old_path, 'new_path': new_path})
            else:
                ops['rename_dirs'].append(op['rename'])
        if 'move' in op:
            old_path = current_path + '/'
            new_path = op['move']['new_path']

            ops['move_dirs'].append({'current_path': old_path, 'new_path': new_path})

    for dir in tree['dirs']:
        final_path = current_path + '/' + dir['dirname']
        get_operations(dir, ops, final_path.replace('//', '/'))


def parse_bmv(data):
    """Parses a BMV string and returns a dictionary of operations

    Args:
        data (str): BMV string

    Returns:
        dict: Dictionary of operations
    """
    parser = Lark.open("bmv.lark", rel_to=__file__, parser='lalr')
    tree = parser.parse(data)
    res = TreeToOperations().transform(tree)

    ops = {'add': [], 'delete': [], 'rename_files': [], 'rename_dirs': [], 'move_files': [], 'move_dirs': []}
    get_operations(res['dir'], ops, res['dir']['dirname'])
    ops['rename_dirs'] = ops['rename_dirs'][::-1]

    return ops



class BmvGenerator:
    def __init__(self):
        pass

    def generate(self, tree):
        """Generates a bmv file from a FileTree

        Args:
            tree (FileTree): the root node of the FileTree

        Returns:
            str: the bmv file contents
        """
        return self._generate_helper(tree, 0)

    def _generate_helper(self, tree, depth):
        tab = ''.join(['\t'] * depth)
        tab_plus_one = tab + "\t"
        name = tree._path() if depth == 0 else tree.name()

        contents = tab + "[" + name + "] {\n"

        for child in tree.children():
            if not child._is_dir():
                contents += tab_plus_one + child.name() + "\n"

        for child in tree.children():
            if child._is_dir():
                contents += self._generate_helper(child, depth + 1)

        return contents + tab + "}\n"



from collections import deque
from .helpers import get_last_index_of_substring


class FileTree:
    def __init__(self, path):
        self.path = path if path[-1] != "/" else path[:-1]
        self._children = []
        self.parent = None
        self.is_directory = False

    def all_paths(self):
        """Genterates every path for every node under a FileTree node

        Returns:
            array: array of string paths
        """
        stack = [self]
        paths = []

        while stack:
            tree = stack.pop()
            paths.append(tree._path() + "/")

            for child in tree.children():
                if child._is_dir():
                    stack.append(child)
                else:
                    paths.append(child._path())

        return paths

    def add_child(self, child):
        """Adds a child FileTree node to the callee FileTree node

        Args:
            child (FileTree): the node to be made a child
        """
        self._children.append(child)
        child.parent = self
        child._set_path(f"{self._path()}/{child.name()}")

        def _update_paths_for_children(tree):
            if tree.children():
                stack = [tree]
                while stack:
                    tree = stack.pop()

                    for tree in tree.children():
                        tree._set_path(f"{tree._parent()._path()}/{tree.name()}")
                        stack.append(tree)

        _update_paths_for_children(child)

    def _set_path(self, path):
        self.path = path

    def children(self):
        """Gets the children of the callee FileTree node

        Returns:
            array: array of FileTree nodes
        """
        return self._children

    def _parent(self):
        return self.parent

    def _path(self):
        return self.path

    def _is_dir(self):
        return self.is_directory

    def name(self):
        """Gets the name of the callee FileTree node

        Is derived from the path. The name is the last substring of the path
        after the last slash.

        For example, if the path is "foo/bar/baz", the name is "baz".

        Returns:
            str: the name of the callee FileTree node
        """
        try:
            last_slash_index = get_last_index_of_substring(self._path(), "/")
            return self._path()[last_slash_index + 1 :]
        except ValueError:
            return self._path()

    def __eq__(self, other):
        stack = deque([self])
        other_stack = deque([other])

        while stack and other_stack:
            tree = stack.pop()
            other_tree = other_stack.pop()

            if len(tree.children()) != len(other_tree.children()):
                return False

            for child in tree.children():
                other_child = None
                for other_child in other_tree.children():
                    if child.name() == other_child.name():
                        break

                if not other_child or child.name() != other_child.name():
                    return False

                if len(child.children()) != len(other_child.children()):
                    return False

                if child._is_dir():
                    stack.append(child)
                    other_stack.append(other_child)

        return len(stack) == 0 and len(stack) == len(other_stack)



from .file_tree_builder import build_from_directory
from .bmv_parser import parse_bmv
from .bmv_generator import BmvGenerator
from os import system
from .operations import perform_adds, perform_deletes, perform_renames, perform_moves


def run(start_path):
    """Runs the full pipeline for bmv.

    Args:
        start_path (str): The directory where you want to run bmv in.

    Returns:
        int: error code

    """
    ft = build_from_directory(start_path)
    bmv_content = BmvGenerator().generate(ft)

    with open("file_tree.bmv", "w") as file:
        file.write(bmv_content)

    system("vim file_tree.bmv")

    with open("file_tree.bmv", "r") as file:
        output = parse_bmv(file.read().strip())

    perform_adds(output["add"])
    perform_deletes(output["delete"])
    perform_renames(output["rename_files"])
    perform_renames(output["rename_dirs"])
    perform_moves(output["move_files"])
    perform_moves(output["move_dirs"])

    perform_deletes(['file_tree.bmv'])

    return 0



from os import listdir
from os.path import isdir
from .file_tree import FileTree

ignore = ['.git', '.github', '.pytest-cache', '__pycache__', '.pytest_cache', 'node_modules', 'tmp']


def build_from_directory(path):
    """Builds a FileTree representation of a directory at a given path

    Args:
        path (str): Path to directory from which a FileTree will be created

    Returns:
        FileTree: The resulting FileTree
    """
    file_tree = FileTree(path)
    ignore_set = set(ignore)

    for entity_name in sorted(listdir(path)):
        new_path = f"{path}/{entity_name}"

        child_file_tree = None
        if isdir(new_path) and entity_name not in ignore_set:
            child_file_tree = build_from_directory(new_path)
            child_file_tree.is_directory = True
        else:
            child_file_tree = FileTree(new_path)

        if entity_name not in ignore_set:
            file_tree.add_child(child_file_tree)

    return file_tree



import unittest
from bulk_mv import build_from_directory, FileTree, parse_bmv
from pathlib import Path


class TestBmvParser(unittest.TestCase):
    def sample1_file_contents(self):
        with open(f"{Path(__file__).parent}/dummy_bmv_files/sample1.bmv") as file:
            return file.read().strip()

    def test_file_tree_visitor(self):
        sample1 = self.sample1_file_contents()
        output = parse_bmv(sample1)

        expected_output = {
            "add": ["web/assets/", "web/pages/images/", "web/pages/images/profile_photos/"],
            "delete": ["web/pages/images/delete_me.jpg"],
            "rename_files": [
                {"old_path": "web/static/main.css", "new_path": "web/static/script.css"},
                {"old_path": "web/static/main.js", "new_path": "web/static/script.js"},
            ],
            "rename_dirs": [],
            "move_files": [
                {"current_path": "web/pages/images/photo.jpg", "new_path": "web/pages/images/profile_photos"}
            ],
            "move_dirs": [{"current_path": "web/static/", "new_path": "web/assets"}],
        }

        self.assertEqual(output, expected_output)
